/* interp.c    interpreter i-code	9 Jan 97

	som-v16 add file i/o 			7 Nov 2005
	som-v16a  optimised speed 		25 Dec 2006 (Christmas!)
	som-v16u  sx-code				1 Mar 2007
 */

#include "compile.h"

#define		PRIVATE 	static

// stack macro
#define   TOS       SS[sp]
#define   NEXT      SS[sp-1]
#define	  ARG		SS[fp-a]
#define   pop2		sp -= 2
#define	  pop1      sp--

//PRIVATE void push(int x) { sp++; SS[sp] = x; }
//PRIVATE int pop(void){ int x; x = TOS; pop1; return x; }
// inline is 4x faster than push/pop functions
#define   push(x) 	sp++; SS[sp] = (x)
#define   pop()		SS[sp]; sp--

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..80   tokstring-som
*/

extern int load_ads;		// base ads of fun "loadfile"

int M[MAXMEM];				// data and code segment
int SS[MAXSS];				// stack segment
int *CS = &M[MAXDS];		// code segment
int DS = MAXSYS;			// current free data segment

PRIVATE FILE *filep[MAXFILE];	// store file pointer
PRIVATE int fp = 1, sp = 0;	// current state of computation
PRIVATE int IP;				// current ip

#ifdef PROF		// --- profiling

int icnt[100], lcnt[10000];

void stat(void){
	icnt[CS[IP] & 255]++;
	if(IP < 10000) lcnt[IP]++;
}

void report(void){
	int i, s;
	s = 0;
	for(i=1; i<EOC; i++)
		if(icnt[i] != 0){
			prCode2(stdout,i);
			printf(" %d\n",icnt[i]);
			s += icnt[i];
		}
	printf("total %d\n",s);
/*
	s = 0;
	for(i=0; i<10000; i++)
		if(lcnt[i] != 0){
			s += lcnt[i];
			printf("%d %d\n",i,lcnt[i]);
		}
	printf("total %d\n",s);
*/
}

#endif

void initfile(void){
	int i;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
}
// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}

#ifdef DEBUG	// ---- instruction tracing

PRIVATE int traceflag = 0;

extern FILE *FO;
extern int lv;

int callstk[200], callstkptr = 0;
void pushcallstk(int x){
	callstkptr++;
	callstk[callstkptr] = x;
}
void popcallstk(void){ callstkptr--; }

void dumpcallstk(void){
	int i;
	printf("call stack :");
	for(i=1; i<=callstkptr; i++)
		printf("%d ",callstk[i]);
	nl();
}

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
	prCode(ip);
	printf(" |");
	a = sp - 10;
	if( a < 0 ) a = 0;

	for(i=a; i<=sp; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}

void dumpDS(void){
	int i;
	for(i=1; i<10; i++)	printf("%d:%d ",i,M[i]);
	printf("\n");
}

#endif

PRIVATE void error(char *mess){
	printf("runtime error: %s at ip %d\n",mess,IP);
	exit(0);
}

// fgets(fp) and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// static alloc memory of size a
PRIVATE int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		error("alloc: out of memory");
	DS += a;
	return b;
}
// exec syscall n
PRIVATE void xsys(int n){
	int a, f;
	char name[NAMELEN], mess[80];

	switch(n){
	case 1: 				// pr int
		a = pop();
		printf("%d",a);
		break;
	case 2: 				// pr char
		a = pop();
		printf("%c",a);
		break;
	case 3: 				// get char
		push(getchar());
		break;
	case 4: 				// gets
		a = pop();
		xgets(stdin,a);
		break;
	case 5:					// fopen fname mode
		a = pop();
		f = pop();			// fname
		strunpack(name,f);
		f = newfp();
		if(f == 0)
			error("cannot open, too many files");
		switch(a){			// mode
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"rw"); break;
		default: error("unknown file mode");
		}
		if(filep[f] == NULL){
			sprintf(mess,"cannot open file %s",name);
			error(mess);
		}
		push(f);			// return file pointer
		break;
	case 6:					// fclose fp
		a = pop();
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:					// fprint fp num
		a = pop();
		f = pop();
		fprintf(filep[f],"%d",a);
		break;
	case 8:					// fprintc fp c
		a = pop();
		f = pop();
		fprintf(filep[f],"%c",a);
		break;
	case 9:
		a = pop();
		push(fgetc(filep[a])); // fgetc fp
		break;
	case 10:				// fgets fp inbuf
		a = pop();
		f = pop();
		xgets(filep[f],a);
		break;
/*
	case 11:				// loadfile in C
		a = pop();
		strunpack(name,a);	// file name in som-string
		loadfile(name);
		break;
*/
	case 12:				// eval from som
		a = pop();
		CS = &M[M[CS_ADS]];	// switch to user space
		eval(a);
		CS = &M[MAXDS];		// switch back
		break;
//	case 13: runflag = 0;  	// halt, this is in eval()
	case 14:
		a = pop();
		push(xalloc(a));
		break;
	case 15:				// exec load from user space
		CS = &M[MAXDS];		// to system
		eval(load_ads);
		CS = &M[M[CS_ADS]];	// switch to user
		break;
	default: error("undefined syscall");
	}
}

void eval(int ref){
//	int limit = 100000;
	int ip, op, runflag;
	int a, b, c, low, hi;

	ip = ref;
	runflag = 1;
    while(runflag ) {
//		if( (limit--) < 0 ) break;
//		if (traceflag) trace();
		IP = ip;
		c = CS[ip];
		op = c & 255;
		a = c >> 8;
		ip++;
		#ifdef PROF
			stat();
		#endif
// ---------- exec ---------

	switch(op){
	case icAdd: NEXT += TOS; pop1; break;
	case icSub: NEXT -= TOS; pop1; break;
	case icMul: NEXT *= TOS; pop1; break;
	case icDiv: NEXT /= TOS; pop1; break;
	case icBand:NEXT &= TOS; pop1; break;
	case icBor: NEXT |= TOS; pop1; break;
	case icBxor:NEXT ^= TOS; pop1; break;
	case icMod: NEXT %= TOS; pop1; break;
	case icEq:  NEXT = NEXT == TOS; pop1; break;
	case icNe:  NEXT = NEXT != TOS; pop1; break;
	case icLt:  NEXT = NEXT <  TOS; pop1; break;
	case icLe:  NEXT = NEXT <= TOS; pop1; break;
	case icGt:  NEXT = NEXT >  TOS; pop1; break;
	case icGe:  NEXT = NEXT >= TOS; pop1; break;
	case icShl: NEXT <<= TOS; pop1; break;
	case icShr: NEXT >>= TOS; pop1; break;

	case icAddv: TOS += ARG; break;
	case icSubv: TOS -= ARG; break;
	case icMulv: TOS *= ARG; break;
	case icDivv: TOS /= ARG; break;
	case icBandv:TOS &= ARG; break;
	case icBorv: TOS |= ARG; break;
	case icBxorv:TOS ^= ARG; break;
	case icModv: TOS %= ARG; break;
	case icEqv:  TOS = TOS == ARG; break;
	case icNev:  TOS = TOS != ARG; break;
	case icLtv:  TOS = TOS <  ARG; break;
	case icLev:  TOS = TOS <= ARG; break;
	case icGtv:  TOS = TOS >  ARG; break;
	case icGev:  TOS = TOS >= ARG; break;
	case icShlv: TOS <<= ARG; break;
	case icShrv: TOS >>= ARG; break;

	case icAddi: TOS += a; break;
	case icSubi: TOS -= a; break;
	case icMuli: TOS *= a; break;
	case icDivi: TOS /= a; break;
	case icBandi:TOS &= a; break;
	case icBori: TOS |= a; break;
	case icBxori:TOS ^= a; break;
	case icModi: TOS %= a; break;
	case icEqi:  TOS = TOS == a; break;
	case icNei:  TOS = TOS != a; break;
	case icLti:  TOS = TOS <  a; break;
	case icLei:  TOS = TOS <= a; break;
	case icGti:  TOS = TOS >  a; break;
	case icGei:  TOS = TOS >= a; break;
	case icShli: TOS <<= a; break;
	case icShri: TOS >>= a; break;

	case icGet: push(ARG); break;
	case icPut: ARG = pop(); break;
	case icLd:	push(M[a]); break;
	case icSt:	M[a] = pop(); break;
	case icLdx: TOS = M[TOS+ARG]; break;
	case icStx:	M[NEXT+ARG] = TOS; pop2; break;
	case icLdy:
		b = TOS + M[a];
		TOS = M[b];
		break;
	case icSty:
		b = NEXT + M[a];
		M[b] = TOS;
		pop2;
		break;

	case icJmp:	ip += a-1; break;
	case icJf:
		b = pop();
		if(b == FALSE) ip += a-1;
		break;
	case icJt:
		b = pop();
		if(b != FALSE) ip += a-1;
		break;
	case icCall:
		b = CS[a] >> 8;		// get lv
		if (sp + b + FUDGE > MAXSS)
			error("stack overflow");
		SS[sp+b] = fp;
		fp = sp + b;
		sp = fp + 1;
		SS[sp] = ip;
		ip = a + 1;
//		pushcallstk(a);	// !!! debug
		break;
	case icRet:
	// return TOS to caller if stack is not empty
		ip = SS[fp+1];
		if ( sp > fp+1 ){
			b = TOS;
			sp = fp - a + 1;
			fp = SS[fp];
			SS[sp] = b;
		}else{
			sp = fp - a;
			fp = SS[fp];
		}
//		popcallstk();		// !!! debug
		break;
	case icEfor:
		ARG++;
		b = ARG <= SS[fp-a+1];	// v <= adj(v)
		push(b);
		break;

	case icInc: ARG++; break;
	case icDec: ARG--; break;
	case icLit:
	case icAds:	push(a); break;
	case icSys:
		if(a == 13) runflag = 0; else xsys(a);
		break;

	case icNot: TOS = TOS == FALSE; break;
	case icCase:
		low = CS[ip] >> 8;
		hi = CS[ip+1] >> 8;
		a = pop();
		if( a < low || a > hi )
			ip += 2;
		else
			ip += a-low+3;
		break;
	case icEnd:	runflag = 0; break;

	default: error("undefined opcode");
    }
// ----- end exec ----
	}
}

